#!/usr/bin/perl

#
# The DaveMP web-based music player interface
#
# Version 1.20
# 
# Written by Jay Kuri Copyright 2007
#
# This work is licensed under the Creative Commons
# Attribution-Noncommercial-Share Alike 2.5 License. To view a copy of this
# license, visit http://creativecommons.org/licenses/by-nc-sa/2.5/ or send a
# letter to Creative Commons, 543 Howard Street, 5th Floor, San Francisco,
# California, 94105, USA.


package Davemp;
use warnings;
use strict;
use Data::Dumper;
use MIME::Base64;
use HTTP::Server::Simple;
use JSON::XS;
use MPD2;
use base qw(HTTP::Server::Simple::CGI);

sub headers {
    my ($self) = shift;
    my $h = shift;
    my %headers =  @$h ; #@_;
    
    $self->config->{'providedauthhash'} = '';
    if (exists($headers{'Authorization'})) {
        if ($headers{'Authorization'} =~ m/Basic (.*)$/) {
            $self->config->{'providedauthhash'} = $1;            
        } 
    } 
}

sub handle_request {
    my ($self, $cgi) = @_;

    my $requestedpath = $cgi->path_info;
    my $prefix;
    my $path;
    (undef, $prefix, $path) = split '/', $requestedpath, 3;
    $prefix ||= '';
    $path ||= 'index.html';
#    print STDERR "prefix/path: $prefix / $path \n";
    if (exists($self->config->{'davemp_http_password'})) {
        my $user = '';
        my $pass = '';
        if ($self->config->{'providedauthhash'} ne '') {
            ($user, $pass) = split /:/, decode_base64($self->config->{'providedauthhash'}), 2;
        }
        if (!($user eq $self->config->{'davemp_http_user'} &&
            $pass eq $self->config->{'davemp_http_password'})) {
            # we return a 401 requesting authorization.
            print "HTTP/1.1 401 Authorization Required\nServer: davemp\n" . 
                  'WWW-Authenticate: Basic realm="DaveMP"' . "\n" .
                  "Content-Type: text/html; charset=iso-8859-1\n\nAuthorization Required";
            return;
        }
    }
    if ($prefix eq 'control') {
        my $command = $path;
        if ($command =~ /^(playlist|library|output|status|nop)/) {
            my $vars = $cgi->Vars;
            map { $vars->{$_} =~ s/\%([A-Fa-f0-9]{2})/pack('C', hex($1))/seg; } keys %$vars;            
            my $result = $self->sendCommand($command, $vars);
            my $jsonstring = to_json($result);
            
            print "HTTP/1.1 200 OK\n";
            #print "Content-type: application/json\n";
            print "Content-type: text/plain\n";
            print "Content-length: " . length($jsonstring) . "\n\n";
            print $jsonstring;
        } else {
            print "HTTP/1.1 404 Not Found\nServer: davemp\nContent-Type: text/html; charset=iso-8859-1\n\nNot Found";
        }
    } else {
        ## this means we are loading themes... IE real files...
        my $themedir = $self->config->{'themeroot'} . "/" . $prefix;
        my $res = 0;
        if ( -d $themedir ) {
            $res = $self->serve_static($themedir . "/" . $path );
        }
        if ($res != 1) {
            print "HTTP/1.1 404 Not Found\nServer: davemp\nContent-Type: text/html; charset=iso-8859-1\n\nNot Found";
        }
    }
}

sub serve_static {
    my ($self, $path) = @_;
        
    ## strip multiple dots followed by a slash.... no backing out of the theme directory, thank you very much
    $path =~ s/\.\.+\///g;
    ##print STDERR $path . "\n";
    
    ## combination mime-type definition and 'allowed file type' list.
    my %mimetypes = ( 'css' => 'text/css',
                      'jpg' => 'image/jpeg',
                      'png' => 'image/png',
                      'gif' => 'image/gif',
                      'js'  => 'application/x-javascript',
                      'json' => 'application/json',
                      'html' => 'text/html',
                    );

    if ($path =~ m|/$| ) {
        $path .= "index.html";
    }                  
    $path =~ /.*\.([^.]+)$/;
    my $mime = $mimetypes{$1};
    if (!$mime) {
        return 0;
    }

    #print STDERR $path . "\n";
    
    if ( -e $path ) {
        my $size = -s $path;
        my $readchunk = ($size > 4096) ? 4096 : $size;
        open STATICFILE, $path || return 0;
        print "HTTP/1.1 200 OK\015\012";
        print "Content-type: " . $mime . "\015\012";
        print "Cache-control: max-age=86400\015\012";
        print "Content-length: " . $size . "\015\012\015\012";
        my $data;
        while (read(STATICFILE, $data, $readchunk)) {
            print $data;
        }
        print $data;
        close STATICFILE;
        return 1;
    }
    return 0;
    
}

sub _davemp {
    my $self = shift;
    
    if (@_) {
        $self->{'_davemp'} = $_[0]
    }
    if (!ref $self->{'_davemp'}) { $self->{'_davemp'} = {}; }
    return $self->{'_davemp'};
}

=head2 setBackend 

Takes a backend object to use with the active DaveMP server

=cut

sub setBackend {
    my ($self, $backend) = @_;
    $self->_davemp->{'backend'} = $backend;
}

sub getBackend {
    my ($self) = shift;
    my $backend = $self->_davemp->{'backend'};
    
    if (!$backend->ensureActive()) {
        print STDERR "Unable to get backend connection\n";
    }
    return($backend);
}

sub print_banner {
    
}

sub sendCommand {
    my ($self, $command, $args) = @_;
    
    my $backend = $self->getBackend();
    my $response = { 
                      'action' => $command,
                      'arguments' => \%{$args},
                   };
    if ($command eq 'nop') {
        $response->{'result'} = "OK";
    } else {
        $response->{'result'} = $backend->sendCommand($command, $args);
    }
    return $response;
}

sub config {
    my $self = shift;
    
    if (@_) {
        $self->_davemp->{'config'} = $_[0]
    } 
    return $self->_davemp->{'config'};
}

1;