package Davemp::Backend::Mpd;

#
# MPD backend for the DaveMP web-based music player interface
#
# Written by Jay Kuri Copyright 2007
#
# This work is licensed under the Creative Commons
# Attribution-Noncommercial-Share Alike 2.5 License. To view a copy of this
# license, visit http://creativecommons.org/licenses/by-nc-sa/2.5/ or send a
# letter to Creative Commons, 543 Howard Street, 5th Floor, San Francisco,
# California, 94105, USA.

use strict;
use MPD2;
use Data::Dumper;
#use Devel::Size qw(size total_size);
use base qw/Class::Accessor::Fast/;

our $VERSION => '0.1';

BEGIN {
    __PACKAGE__->mk_accessors(qw/config mpdhandle state currentplaylist library/);
}

sub new {
	my($class, $config) = @_;
		
	my $self = { 
	               config => {},
	               state => {},
	               mpdhandle => undef,
	               currentplaylist => [],
	               library => undef,
	            };
	        
	 bless $self, $class;
	 
	 if (ref($config) eq 'HASH') {
	     $self->config($config);
	 }
	 
	 ## set up some defaults if they weren't already set.
     $self->config->{'mpd_librarytimeout'} ||= 3600;
	 $self->config->{'mpd_metaindex1'} ||= 'artist';
     $self->config->{'mpd_metaindex2'} ||= 'album';
     $self->config->{'mpd_preloadlibrary'} ||= 'true';
      
	 $self->state->{'libraries'}{'default'} = { 
                                                  'name' => 'default', 
                                                  'description' => 'Main MPD Library',
                                                  'metadata' => [
                                                                   'album',
                                                                   'artist',
                                                                   'title',
                                                                   'genre',
                                                                   'track',
                                                                   'path1',
                                                                   'path2',
                                                                   'filename'
                                                                ],
                                                  'commands' => [
                                                                   'info',
                                                                   'list',
                                                                   'reload',
                                                                   'trackinfo',
                                                                   'status'
                                                                ]
                                              };
     
	 $self->mpdhandle(MPD2->new($self->config));
	 
	 
	 if ($self->config->{'mpd_preloadlibrary'} ne 'false') {
         $self->loadLibrary();
     }
	 
	 return $self;
}

sub ensureActive {
    my $self = shift;
    
    if (!$self->mpdhandle->is_connected) {
        if ($self->mpdhandle->connect()) {
            $self->loadLibrary();
            $self->loadStatus();
            $self->loadOutputStatus();
        } else {
            return undef;
        }
    }
    return $self;
}

sub sendCommand {
    my ($self, $command, $args) = @_;
    
    $self->state->{'now'} = time();
    if ($command =~ m|^playlist/(.*)$| ) {
        return $self->processPlaylistCommand($1, $args);
    } elsif ($command =~ m|^library/(.*)$| ) {
        return $self->processLibraryCommand($1, $args);
    } elsif ($command =~ m|^output/(.*)$| ) {
        return $self->processOutputCommand($1, $args);
    } elsif ($command =~ m|^status$| ) {
        return $self->globalStatus($args);
    } else {
        return undef;
    }
}

sub processPlaylistCommand {
    my ($self, $command, $args) = @_;
    
    ## first part of the playlist command is always the playlist name - we only ever have one in MPD 
    ## so we just chop it off.  all playlists = 'active' for us.
    $command =~ m|^([^/]+)/(.*)$|; 
    
    my $playlistname = $1;    
    $command = $2;

    if ($command eq 'status') {
        $self->loadStatus();
        my $result = { 
                        'name' => 'active', 
                        'description' => 'Active Playlist',
                        'currentposition' => $self->state->{'status'}{'song'},
                        'length' => $self->state->{'status'}{'playlistlength'},
                        'key' => $self->state->{'status'}{'playlist'},
                     };
        $result->{'activetrackmetadata'} = {};
        foreach my $key (keys %{$self->state->{'status'}{'currenttrack'}}) {
            if ($key eq 'time') { 
                $result->{'activetrackmetadata'}{'length'} = $self->state->{'status'}{'currenttrack'}{$key};
            } elsif ($key eq 'pos') {
                $result->{'activetrackmetadata'}{'position'} = $self->state->{'status'}{'currenttrack'}{$key};
            } else {
                $result->{'activetrackmetadata'}{lc($key)} = $self->state->{'status'}{'currenttrack'}{$key};   
            }
        }
        return $result;
    } elsif ($command eq 'get') {
        my $result = $self->mpdhandle->send_grouped_command('playlistinfo', qr/^(file)/);
        map { $_->{'id'} = $_->{'file'} ; $_->{'position'} = $_->{'pos'} ; delete $_->{'pos'}} @{$result};
        return $result;
    } elsif ($command eq 'removetrack') {
        if (exists($args->{'position'})) {
            $self->mpdhandle->send_simple_command('delete ' . $args->{'position'});
            return 'OK';
        } else {
            return undef;
        }
    } elsif ($command eq 'addtrack') {
        if (exists($args->{'trackid'})) {
            my ($library, $trackid) = split(':', $args->{'trackid'}, 2);  # we ignore library since there is only one in mpd.
            if ($self->mpdhandle->send_simple_command('add "' . $trackid . '"')) {
                $self->loadStatus(1);
                if (exists($args->{'position'})) { 
                    my $origposition = $self->state->{'status'}{'playlistlength'} - 1;
                    $self->mpdhandle->send_simple_command('move ' . $origposition . ' ' . $args->{'position'});
                }
                return $self->processLibraryCommand($library . "/" . "trackinfo", { id => $trackid });
            } else {
                return undef;
            }
        } else {
            return undef;
        }
    } elsif ($command eq 'movetrack') {
        if (exists($args->{'oldposition'}) && exists($args->{'newposition'})) {
            return $self->mpdhandle->send_simple_command('move ' . $args->{'oldposition'} . ' ' . $args->{'newposition'});
        } else {
            return undef;
        }
    } elsif ($command eq 'copy') {
        ## since the only real playlist in mpd is 'active' - copying means either copying to active - loading, or 
        ## copying from active - saving.  so that's how we interpret it.
        if (exists($args->{'playlist'})) {
            if ($playlistname eq 'active') {
                $self->mpdhandle->send_simple_command('clear');
                return $self->mpdhandle->send_simple_command('load "'. $args->{'playlist'} . '"');
            } else {
                return $self->mpdhandle->send_simple_command('save "' . $args->{'playlist'} . '"');
            }
        }
    }
    
}

sub processLibraryCommand {
    my ($self, $command, $args) = @_;
    
    my $libraryname = 'active';
    if ($command =~ /^(list)$/) {
        $command = "/Rootlist";
    } else {
        ## first part of the library command is always the library name - we only ever have one in MPD 
        ## so we just chop it off.  all playlists = 'active' for us.
        $command =~ m|^([^/]+)/(.*)$|; 

        $libraryname = $1;
        $command = $2;
    }
    
    if ($command eq 'status') {
        return { status => 'active' };
    } elsif ($command eq '/Rootlist') {
        return [ $self->state->{'libraries'}{'default'} ];
    } elsif ($command eq 'reload') {
        $self->mpdhandle->send_simple_command('update');
        return 'OK';
    } elsif ($command eq 'info') {
        return $self->state->{'libraries'}{$libraryname};
    } elsif ($command eq 'trackinfo') {
        my $res = $self->processLibraryCommand($libraryname . '/list', { 'fields' => 'all', 'search' => 'id:=' . $args->{'id'}});
        if ($res) {
            return $res->[0];
        } else {
            return undef;
        }
    } elsif ($command eq 'settrackinfo') {
        # this method always fails on the mpd backend, since mpd does not edit files.
        ## replacement metadata / additional metadata _could_ be saved by the backend in it's own
        ## database - but that is not built yet - and may not be.  depends on interest.
        return 'FAILED';
    } elsif ($command eq 'list') {
        ### UGGG-LEEE.  But what can you do. The metadata searching in mpd is not so hot... 
        ### so we make do by working directly with the full library.  Woot.  At least we cache
        ### it, right?
        
        $self->loadLibrary();
        
        my $matcheditems = [];
        my $fieldstoreturn = [];
        if ($args->{'fields'} && $args->{'fields'} ne 'all') {
            @{$fieldstoreturn} = split(',', $args->{'fields'});
        }
#        my @searchfields;
        my %searchterms;
#        my $mashterm;
        my $matchesrequired = undef;
        if ($args->{'search'}) {
            my @searchfields = split(/\|/, $args->{'search'});
            # print STDERR Dumper(\@searchfields);            
            foreach my $search (@searchfields) {
                if ($search =~ m/^([^:]+):(.*)/) {
                    my $key = $1;
                    my $val = $2;
                    if ($val =~ /^=(.*)$/) {
                        $val = '^' . $1 . '$';
                        $searchterms{$key} = $1;
                    } else {
                        $searchterms{$key} = qr/$val/;
                    }
#                    $mashterm = $val;
                    $matchesrequired++;
                }
            }
        }
        
        my $idx1 = $self->config->{'mpd_metaindex1'};
        my $idx2 = $self->config->{'mpd_metaindex2'};
#        $mashterm =~ s/[\^\$]/:/g;
#        $mashterm = qr/$mashterm/;
#        print STDERR Dumper($mashterm);



        my @firstkeys;
        if (exists($searchterms{$idx1})) {
            #print STDERR "using index 1 ";
            if (ref($searchterms{$idx1}) ne 'Regexp') {
                push @firstkeys, $searchterms{$idx1};
            } else {
                #print STDERR "as Regex";
                push @firstkeys, grep { $_ =~ $searchterms{$idx1} } sort keys %{$self->library};
            }
            #print STDERR "\n";
        } else {
            push @firstkeys, sort keys %{$self->library};
        }
        
        # optimization for requests for just one of the two indexes we use.
        if ($#{$fieldstoreturn} == 0) {
            if ($fieldstoreturn->[0] eq $idx1) {
                push @{$matcheditems}, map { { $idx1 => $_ } } @firstkeys;
                #print STDERR $idx1 . " optimization\n";
                return($matcheditems);
            }
        }
                
        my $tosearch = [];
        my @keystosearch;
        foreach my $key (@firstkeys) {
            if (exists($searchterms{$idx2})) {
                #print STDERR "using index 2 ";
                if (ref($searchterms{$idx2}) ne 'Regexp') {
                    push @{$tosearch}, $self->library->{$key}->{$searchterms{$idx2}};
                    push @keystosearch, $searchterms{$idx2};
                } else {
                    #print STDERR "as regex";
                    foreach my $key2 (sort keys %{$self->library->{$key}}) {
                        if ($key2 =~ $searchterms{$idx2}) {
                            push @{$tosearch}, $self->library->{$key}->{$key2};
                            push @keystosearch, $key2;
                        } 
                    }
                }
                #print STDERR "\n";
            } else {   
                foreach my $key2 (sort keys %{$self->library->{$key}}) {
                    push @{$tosearch}, $self->library->{$key}->{$key2};
                    push @keystosearch, $key2;
                }
            }
        }
        if ($#{$fieldstoreturn} == 0 && $fieldstoreturn->[0] eq $idx2) {
            push @{$matcheditems}, map { { $idx2 => $_ } } @keystosearch;
            #print STDERR $idx2 . " optimization\n";
            return($matcheditems);
        } else {
            #print STDERR Dumper($matcheditems);
            
            push @{$matcheditems}, @{$self->searchBatch(\%searchterms, $fieldstoreturn, $tosearch)};
        }
        #print STDERR Dumper($matcheditems);
        return($matcheditems);
        
    }
    
}

sub searchBatch {
    my ($self, $searchterms, $fieldstoreturn, $lists) = @_;    
    
    my $matchesrequired = keys %$searchterms;
    my $matcheditems = [];
    my %foundkeys;
    
    foreach my $list (@{$lists}) {
        foreach my $libraryitem (@{$list}) {
            my $matchcount = 0;
            if ($matchesrequired ) {
                foreach my $key (keys %{$searchterms}) {
                    my $search = $searchterms->{$key};
                    if (ref($search)) {
                        if ($libraryitem->{$key} =~ $search) {
                            $matchcount++;
                        }
                    } else {
                        if ($libraryitem->{$key} eq $search) {
                            $matchcount++;
                        }
                    }
                }
            }
            if ($matchesrequired == undef || $matchcount == $matchesrequired) {
                my $key;
                if ($#{$fieldstoreturn} >= 0) {
                    $key = join '', map { $libraryitem->{$_} } @{$fieldstoreturn};
                    if (!exists($foundkeys{$key})) {
                        $foundkeys{$key} = 1;
                        my %item;
                        my $fieldmatchcount = 0;
                        foreach my $field (@{$fieldstoreturn}) {
                            if (exists($libraryitem->{$field})) {
                                $item{$field} = $libraryitem->{$field};
                                $fieldmatchcount++;
                            }
                        }
                        #my %item = map { $_ => $libraryitem->{$_} } @{$fieldstoreturn};
                        if ($fieldmatchcount) { 
                            push @{$matcheditems}, \%item;
                        }
                    }
                } else {
                    push @{$matcheditems}, $libraryitem;
                }
            }
        }
    }
    # print STDERR $#{$matcheditems} . " matched\n";

    return($matcheditems);

}

sub processOutputCommand {
    my ($self, $command, $args) = @_;
    
    my $outputname = 'active';
    if ($command =~ /^(list|setactive)$/) {
        $command = $1;
    } else {
        $command =~ m|^([^/]+)/(.*)$|;
        $outputname = $1;
        $command = $2;
    }
    if ($outputname eq 'active') {
        $outputname = $self->state->{'activeoutput'};
    }
    
    #print STDERR "output: $outputname\n";
    
    if ($command eq 'status') {
        $self->loadStatus();
        $self->loadOutputStatus();
        my $result = { 
                        gain => $self->state->{'status'}{'volume'},
                        crossfade => $self->state->{'status'}{'xfade'},
                        'repeat' => ($self->state->{'status'}{'repeat'} == 0) ? 'off' : 'playlist',
                        'random' => ($self->state->{'status'}{'random'} == 0) ? 'off' : 'on',
                     };

        if ($self->state->{'status'}{'state'} eq 'play') {
            $result->{state} = 'playing';
        } elsif ($self->state->{'status'}{'state'} eq 'pause') {
            $result->{state} = 'paused';
        } else {
            $result->{state} = 'stopped';
        }
        ($result->{'trackelapsed'}, $result->{'tracklength'}) = split ':', $self->state->{'status'}{'time'};
        
        #if ($self->state->{'status'}{'time'} =~ m|(^\d+):(\d+)|o) {
        #    $result->{'trackelapsed'} = $1;
        #    $result->{'tracklength'} = $2;
        #}
        
        $result->{'info'} = $self->state->{'outputs'}{$outputname};
        
        return $result;

    } elsif ($command eq 'pause') {

        $self->mpdhandle->send_simple_command('pause');
        return 'OK';

    } elsif ($command eq 'stop') {

        $self->mpdhandle->send_simple_command('stop');
        return 'OK';

    } elsif ($command eq 'next') {

        $self->mpdhandle->send_simple_command('next');
        return 'OK';

    } elsif ($command eq 'previous') {

        $self->mpdhandle->send_simple_command('previous');
        return 'OK';
        
    } elsif ($command eq 'play') {
        
        $self->loadStatus();
        if (!exists($args->{'track'}) && !exists($args->{'seek'})) {
            $self->mpdhandle->send_simple_command('play');
        } else {
            my $track = $self->state->{'status'}{'song'};
            $track = $args->{'track'} if exists($args->{'track'});
            my $seek = $args->{'seek'} || 0;
            $self->mpdhandle->send_simple_command('seek ' . $track . ' ' . $seek);
            if ($self->state->{'status'}{'state'} != 'play') {
                $self->mpdhandle->send_simple_command('play');
            }
        }
        return $self->processOutputCommand( $outputname . '/status');
    
    } elsif ($command eq 'setgain') {
        
        if (exists($args->{'level'}) &&  0 <= $args->{'level'} && $args->{'level'} <= 100 ) {
            return $self->mpdhandle->send_simple_command("setvol " . $args->{'level'});
        } else {
            return undef;
        }
    } elsif ($command eq 'random') {
        my $newstate = $self->state->{'random'} ? 0 : 1;
        if (exists($args->{'state'})) {
            $newstate = $args->{'state'};
        }
        return $self->mpdhandle->send_simple_command("random " . $newstate);
    } elsif ($command eq 'repeat') {
        my $newstate = ($self->state->{'repeat'} == 1) ? 0 : 1;
        if (exists($args->{'state'})) {
            $newstate = $args->{'state'};
        }
        return $self->mpdhandle->send_simple_command("repeat " . $newstate);
    } elsif ($command eq 'crossfade') {
        if ($args->{'time'} && 0 <= $args->{'time'} && $args->{'time'} <= 30 ) {
            return $self->mpdhandle->send_simple_command("crossfade " . int($args->{'time'}));
        } else {
            return undef;
        }
    } elsif ($command eq 'list') {
        ## for now all the specific output commands are nulled out - someone should fill 
        ## these in with mpd's multi-output stuff.  but not me, tonight.
        $self->loadOutputStatus();
        my @results;
        foreach my $output (keys %{$self->state->{'outputs'}}) {
            push @results, $self->state->{'outputs'}{$output};
        }
        return \@results;
    } elsif ($command eq 'enable') {
        $self->mpdhandle->send_simple_command('enableoutput 0');
        $self->loadOutputStatus();
        return 'OK';
    } elsif ($command eq 'disable') {
        $self->mpdhandle->send_simple_command('disableoutput 0');
        $self->loadOutputStatus();
        return 'OK';
    } elsif ($command eq 'setactive') {
        $self->state->{'activeoutput'} = $args->{'id'};
        return 'OK';
    } elsif ($command eq 'setplaylist') {
        if (exists($args->{'playlist'})) {
            if ($args->{'playlist'} ne 'active') {
                return $self->processPlaylistCommand("active/copy", { playlist => $args->{'playlist'}});
            } else {
                return 'OK';
            }
        } else {
            return undef;
        }
    }
    
    return(undef);
}

sub loadOutputStatus {
    my ($self, $force) = @_;
    
    if ( $self->state->{'outputstatusloaded'} != $self->state->{'now'} || $force ) {
    
        my $res = $self->mpdhandle->send_grouped_command('outputs', qr/outputid/);
    
        for my $outputrow (@{$res}) {
            # $outputrow->{'outputname'} =~ s/\s//g;
            my $output = {
                    'type' => 'direct', 
                    'name' => $outputrow->{'outputid'},
                    'description' => $outputrow->{'outputname'},
                    'enabled' => (($outputrow->{'outputenabled'} == 0) ? 'off' : 'on'),
                    'commands' => [
                                    'status',
                                    'play',
                                    'pause',
                                    'stop',
                                    'next',
                                    'previous',
                                    'setgain',
                                    'crossfade',
                                    'enable',
                                    'disable'
                                  ],
            };
            $self->state->{'outputs'}{$output->{'name'}} = $output;
        }
        $self->state->{'outputstatusloaded'} = $self->state->{'now'};
    }
}

sub loadStatus {
    my ($self, $force) = @_;
    
    if ( $self->state->{'statusloaded'} != $self->state->{'now'} || $force ) {
        $self->state->{'status'} = $self->mpdhandle->send_short_command("status");
        $self->state->{'status'}{'currenttrack'} = $self->mpdhandle->send_short_command("currentsong");
        $self->state->{'status'}{'currenttrack'}{'id'} =  $self->state->{'status'}{'currenttrack'}{'file'};
        $self->state->{'statusloaded'} = $self->state->{'now'};
    }
}

sub loadLibrary {
    my $self = shift;
    if (!($self->library) || 
          ($self->state->{'librarychecked'} < ($self->state->{'now'} - $self->config->{'mpd_librarytimeout'}))) {
    
        #print STDERR "Library timed out" . $self->state->{'librarychecked'} . "\n";
        my $stats = $self->mpdhandle->send_command("stats");
        
        if (!exists($self->state->{'mpd_dbupdate'}) || $self->state->{'mpd_dbupdate'} < $stats->{'db_update'}) {
        
            #print STDERR "Loading Library: " . $stats->{'db_update'} . "\n";
        
            $self->state->{'mpd_dbupdate'} = $stats->{'db_update'};
            my $idx1 = $self->config->{'mpd_metaindex1'};
            my $idx2 = $self->config->{'mpd_metaindex2'};
            my $usespathinfo = 0;
            if ($idx1 =~ /path/ || $idx2 =~/path/) {
                $usespathinfo = 1;
            }
            
            $self->library({});
            my $library = $self->mpdhandle->send_command("listallinfo");
            foreach my $item (@{$library}) {
                if (exists($item->{'file'})) {
                    # print STDERR "mash: " . $item->{'mash'} . "\n";
                    $item->{'id'} = $item->{'file'};
                    $item->{'id'} =~ m|^(.*)/([^\/]+)\.([^\.]+)$|o;
                    $item->{'filename'} = $2 . '.' . $3;
                    my @parts;
                    if ($usespathinfo) {
                        $item->{'dir'} = $1;
                        $item->{'type'} = lc($3);
                        @parts = split '/', $1; 
                        foreach my $part (0..$#parts) {
                            my $foo = "path" . ($part+1);
                            $item->{$foo} = $parts[$part];
                            #print STDERR $foo . " = " . $parts[$part] . "\n";
                        }
                    }
                    if (!exists($item->{'artist'})) {
                        if ($parts[0] ne '') {
                            $item->{'artist'} = $parts[0];
                        } else {
                            $item->{'artist'} = "Unknown";
                        }
                    }
                    if (!exists($item->{'album'})) {
                        if ($parts[1] ne '') {
                            $item->{'album'} = $parts[1];
                        } else {
                            $item->{'album'} = "Unknown";
                        }
                    }
                    if (!exists($item->{'title'})) {
                        $item->{'title'} = $item->{'filename'};
                    }
                    delete $item->{'file'};
                    if (!exists($self->library->{$item->{$idx1}})) {
                        $self->library->{$item->{$idx1}} = {};
                    }
                    if (!exists($self->library->{$item->{$idx1}}->{$item->{$idx2}})) {
                        $self->library->{$item->{$idx1}}->{$item->{$idx2}} = [];
                    }
                    push @{$self->library->{$item->{$idx1}}->{$item->{$idx2}}}, $item;
                    #push @{$self->library}, $item;
                }
            }
            #print STDERR "Total Size of Library: " . total_size($self->library) . "\n";
            # $self->library($library);
        }
        $self->state->{'librarychecked'} = $self->state->{'now'};
        
    }
}

sub globalStatus {
    my ($self, $command) = @_;
 
    my $result = {};
    $result->{'output'} = $self->processOutputCommand("active/status");
    $result->{'playlist'} = $self->processPlaylistCommand("active/status");
    $result->{'library'} = $self->processLibraryCommand("active/status");
    
    return($result);
}

sub addat {
    my ($self, $command) = @_;
    
    if ($#_ == 1) {
        $self->send_simple_command("add", $_[0]);
        my $status = $self->send_short_command("status");
        if ($status->{'playlistlength'} != $_[1]) {
            my $movecmd = "move " . $status->{'playlistlength'} . " " . $_[1];
            $self->send_simple_command($movecmd);
        }
        my $result = $self->send_grouped_command("playlistinfo " . $_[1]);
        return $result->[0];
    } else {
        return undef;
    }
}


1;