package MPD2;

#
# MPD2 - a perl module for access to the Mpd music player daemon.
#
# Written by Jay Kuri  Copyright 2007
#
# This work is licensed under the Creative Commons Attribution 2.5 License. To
# view a copy of this license, visit
# http://creativecommons.org/licenses/by/2.5/ or send a letter to Creative
# Commons, 543 Howard Street, 5th Floor, San Francisco, California, 94105,
# USA.



use strict;
use IO::Socket;
use Data::Dumper;
use base qw/Class::Accessor::Fast/;

our $VERSION => '0.1';

BEGIN {
    __PACKAGE__->mk_accessors(qw/config state _socket outputs commands error/);
}

sub new {
	my($class, $config) = @_;
		
	my $self = { 
	           config => {},
	           state => {},
	           outputs => {},
	           commands => {},
	           error => undef,
	           _socket => undef
	        };
	        
	 bless $self, $class;
	 
	 if (ref($config) eq 'HASH') {
	     $self->config($config);
	 }
	 
	 $self->config->{mpd_port} ||= $ENV{'MPD_PORT'} || '6600';
	 $self->config->{mpd_host} ||= $ENV{'MPD_HOST'} || '127.0.0.1';
	 $self->config->{mpd_timeout} ||= 30;
	 
	 return $self;
}

sub connect {
    my $self = shift;
    my $socket = new IO::Socket::INET( 
                                     Proto => 'tcp',
                                     PeerAddr => $self->config->{'mpd_host'},
		                             PeerPort => $self->config->{'mpd_port'},
	                               );
#	print STDERR "Connecting\n";
    $self->_socket($socket);
	if ($self->_socket && $self->_socket->connected) {
	    if(<$socket> =~ /^OK MPD (.+)$/)
    	{
    		$self->state->{'version'} = $1;
    		$self->state->{'connected'} = 1;
    	}
	    if ($self->config->{'mpd_password'}) {
	        my $res = $self->send_simple_command("password " . $self->config->{'mpd_password'});
	        if ($res !~ /OK/) {
	            $self->disconnect;
	            return undef;
	        }
	    } 
	    return 1;
	    #$self->status();
	    #$self->get_outputs();
	    #$self->get_commands();
	} else {
	    return undef;
	}
}

sub disconnect {
    my $self = shift;
    
    $self->send_simple_command("close");
    $self->_socket->close;
    $self->state->{'connected'} = 0;
}

sub status {
    my $self = shift;
    my $status = $self->send_short_command("status");
    if ($status) {
        $self->state->{'status'} = $status;
        return $self->state->{'status'};
    } else {
        return undef;
    }
}

sub is_connected {
    my $self = shift;
    my $now = time();
    if ($self->state->{'connected'} && $self->_socket->connected &&
        $self->state->{'lastcommand'} < $now - ($self->config->{'mpd_timeout'})) {
        ## unfortunately, our link can disappear without us knowing it... so every so often 
        ## we have to ping it to check.  We do this if we haven't sent a command in mpd_timeout seconds.
        if ($self->send_simple_command("ping") ne 'OK') {
            $self->state->{'connected'} = 0;
        }
    }
    return $self->state->{'connected'};
}

sub _send_raw_command {
    my ($self, $command) = @_;
    
    if (!$self->_socket || !$self->_socket->connected) {
        $self->connect();
    }
    #print STDERR "sending command: " . $command . " at " . time() . "\n";
    print {$self->_socket} $command . "\n";
    $self->state->{'lastcommand'} = time();
}

sub _record_error {
    my ($self, $errormsg) = @_;
    
    if ($errormsg =~ /^ACK \[(\d+)\@(\d+)\] {(.*)} (.+)$/) {
        $self->error({ error_id => $1,
                       error_command_id => $2,
                       error_command => $3,
                       error_message => $4 });
	}
}

sub clear_error {
    my $self = shift;
    
    $self->error(undef);
}

=head2 send_simple_command $command

Sends a command that expects a single line response and returns the response directly to the caller.

=cut 

sub send_simple_command {
    my ($self, $command) = @_;
    
    $self->_send_raw_command($command);
    my $socket = $self->_socket;
    my $res = <$socket>;
    chomp $res;
    #print STDERR $res . "\n";
    if ($res =~ /^ACK/ ) {
        $self->_record_error($res);
        return undef;
    }
    return $res;
}


=head2 send_long_command $command

sends a command that you expect a long reply to.  Returns an array to the caller containing the full response from the server.

=cut 

sub send_long_command {
    my ($self, $command) = @_;
    
    $self->_send_raw_command($command);
    my $socket = $self->_socket;
    my @results;
	while(<$socket>) 
	{
		chomp;
		#print STDERR "foo:" . $_ . "\n";
		if(/^ACK /) {
		    $self->_record_error($_);
			return undef;
		}
		last if /^OK/;
		push @results, $_;
	}
	#print STDERR join("\n", @results);
    return \@results;
}

=head2 send_command $command

sends a normal command that expects a small batch of lines in response, returns a hashref.  This should only be used
for routines that do not return duplicate keys, obviously.

=cut 

sub send_short_command {
    my ($self, $command) = @_;
    
    my %results;
    my $res = $self->send_long_command($command);
    if (!$res) {
        return $res;
    } 
    my @lines = @{$res};
    if ($#lines >= 0) {
	    foreach my $line (@lines) 
	    {
		    if ( $line =~ /^(.[^:]+)\s*:\s*(.+)$/ ) {
		        $results{lc($1)} = $2;
		    }
	    }
        return \%results;
	} else {
	    return undef;
	}
}

=head2 send_grouped_command $command $delimeters

Routine to run commands that return data in groups.  Separates one group from the next by paying attention to what data
it has already received for a particular hash.  If it sees a key that it already has for the current group, it triggers a new record.  
$delimeters is a regex to match on keys, which is used to separate one record from the next, in addition to the 
key duplicate method above.  This allows the grouped routine to cope with records that are sometimes 'directory' and sometimes 'file'.
Returns an array of hashes.

=cut

sub send_grouped_command {
    my ($self, $command, $recorddelimiters) = @_;

    my @results;
    
    my $res = $self->send_long_command($command);
    if (!$res) {
        return $res;
    } 
    my @lines = @{$res};
    if ($#lines >= 0) {
        my $subgroup = undef;
	    foreach my $line (@lines) {
            if ( $line =~ /^(.[^:]+)\s*:\s*(.+)$/ ) {
    		    my $key = lc($1);
    		    my $value = $2;
    		    if ($subgroup && (exists($subgroup->{$key}) || ($recorddelimiters && $key =~ $recorddelimiters))) {
    		        push @results, $subgroup;
    		        $subgroup = {};
    		    }
    		    $subgroup->{$key} = $value;
    		}
        }
        push @results, $subgroup;
        return \@results;
    } else {
        return undef;
    }
}

=head2 send_command $command @args

The big daddy of the MPD2 perl lib.  This takes a command and arguments as a list, and processes it, returning the 
appropriate data structure for the command, or undef on failure.

=cut 

our $mpdcommands = {
    #    Admin Commands 
    'disableoutput' =>    { maxargs => 1, minargs => 1, type => 'simple' },
    'enableoutput' =>     { maxargs => 1, minargs => 1, type => 'simple' },
    'kill' =>             { maxargs => 0, type => 'simple' },
    'outputs' =>          { maxargs => 0, type => 'sub', 'sub' => 'outputs' },
    'update' =>           { maxargs => 1, type => 'short' },
    #    Database Commands 
    'find' =>             { maxargs => 2, minargs => 2, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/ },
    'list' =>             { maxargs => 3, minargs => 1, type => 'grouped'  },
    'listall' =>          { maxargs => 1, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/ },  # pick up here.
    'listallinfo' =>      { maxargs => 1, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/ },
    'lsinfo' =>           { maxargs => 1, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/ },
    'search' =>           { maxargs => 2, minargs => 2, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/ },
    #    Playlist Commands 
    'add' =>              { maxargs => 1, minargs => 1, type => 'simple' },
    'addid' =>            { maxargs => 1, minargs => 1, type => 'simple' },
    'addat' =>            { maxargs => 2, minargs => 2, type => 'sub', 'sub' => 'addat' },
    'clear' =>            { maxargs => 0, type => 'simple' },
    'currentsong' =>      { maxargs => 0, type => 'short' },
    'delete' =>           { maxargs => 1, minargs => 1, type => 'simple' },
    'deleteid' =>         { maxargs => 1, minargs => 1, type => 'simple' },
    'load' =>             { maxargs => 1, minargs => 1, type => 'simple' },
    'move' =>             { maxargs => 2, minargs => 2, type => 'simple' },
    'moveid' =>           { maxargs => 2, minargs => 2, type => 'simple' },
    # 'playlist' =>         { maxargs => 3, type => 'short' },  # deprecated.
    'playlistinfo' =>     { maxargs => 1, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/ },
    'playlistid' =>       { maxargs => 1, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/  },
    'plchanges' =>        { maxargs => 1, minargs => 1, type => 'grouped', 'delimiters' => qr/(file|directory|playlist)/  },
    'plchangesposid' =>   { maxargs => 1, minargs => 1, type => 'grouped' },
    'rm' =>               { maxargs => 1, minargs => 1, type => 'simple' },
    'save' =>             { maxargs => 1, minargs => 1, type => 'simple' },
    'shuffle' =>          { maxargs => 0, type => 'simple' },
    'swap' =>             { maxargs => 2, minargs => 2, type => 'simple' },
    'swapid' =>           { maxargs => 2, minargs => 2, type => 'simple' },
    #    Playback Commands 
    'crossfade' =>        { maxargs => 1, minargs => 1, type => 'simple' },
    'next' =>             { maxargs => 0, type => 'simple' },
    'pause' =>            { maxargs => 1, type => 'simple' },
    'play' =>             { maxargs => 1, type => 'simple' },
    'playid' =>           { maxargs => 1, type => 'simple' },
    'previous' =>         { maxargs => 0, type => 'simple' },
    'random' =>           { maxargs => 1, type => 'simple' },
    'repeat' =>           { maxargs => 1, type => 'simple' },
    'seek' =>             { maxargs => 2, minargs => 2, type => 'simple' },
    'seekid' =>           { maxargs => 2, minargs => 2, type => 'simple' },
    'setvol' =>           { maxargs => 1, minargs => 1, type => 'simple' },
    'stop' =>             { maxargs => 0, type => 'simple' },
    #    Miscellaneous Commands 
    'clearerror' =>       { maxargs => 0, type => 'simple' },
    'close' =>            { maxargs => 0, type => 'simple' },
    'commands' =>         { maxargs => 0, type => 'grouped' },
    'notcommands' =>      { maxargs => 0, type => 'grouped' },
    'password' =>         { maxargs => 1, minargs => 1, type => 'simple' },
    'ping' =>             { maxargs => 1, type => 'simple' },
    'stats' =>            { maxargs => 0, type => 'short' },
    'status' =>           { maxargs => 0, type => 'short' },
    'fullstatus' =>       { maxargs => 0, type => 'sub', 'sub' => 'fullstatus'}
};


sub send_command {
    my ($self) = shift;
    my ($command) = shift;
    
    #print STDERR $command . " - " . Dumper($commands) . "\n";
    if (!exists($mpdcommands->{$command}) || (exists($mpdcommands->{$command}{minargs}) && $#_ + 1 < $mpdcommands->{$command}{minargs})) {
        return undef;
    }
    
    ## figure out the right function to call for the command in question, and call it.
    ## build up args
    my $mpdcommand = $command;
    my $arg = 0;
    while (($arg < $mpdcommands->{$command}{maxargs}) && $arg <= $#_ ) {
        
        $mpdcommand .= ' "' . $_[$arg++] . '"';
    }
    
    #print STDERR "preparing to send " . $commands->{$command}{'type'} . " command: " . $mpdcommand . "\n";
    if ($mpdcommands->{$command}{'type'} eq 'simple') {
        return $self->send_simple_command($mpdcommand);
    } elsif ($mpdcommands->{$command}{'type'} eq 'short') {
        return $self->send_short_command($mpdcommand)
    } elsif ($mpdcommands->{$command}{'type'} eq 'grouped') {
        return $self->send_grouped_command($mpdcommand, $mpdcommands->{$command}{'delimiters'});
    } elsif ($mpdcommands->{$command}{'type'} eq 'long') {
        return $self->send_long_command($mpdcommand)
    } elsif ($mpdcommands->{$command}{'type'} eq 'sub') {
        my $routine = $mpdcommands->{$command}{'sub'};
        return $self->$routine($command, @_);
    } else {
        return undef;
    }
}

sub fullstatus {
    my ($self, $command) = @_;
    
    my $hash = $self->send_short_command("status");
    $hash->{song} = $self->send_short_command("currentsong");
    return $hash;
}

sub addat {
    my ($self, $command) = @_;
    
    if ($#_ == 1) {
        $self->send_simple_command("add", $_[0]);
        my $status = $self->send_short_command("status");
        if ($status->{'playlistlength'} != $_[1]) {
            my $movecmd = "move " . $status->{'playlistlength'} . " " . $_[1];
            $self->send_simple_command($movecmd);
        }
        my $result = $self->send_grouped_command("playlistinfo " . $_[1]);
        return $result->[0];
    } else {
        return undef;
    }
}


1;