/*
    mootools based dragging controller
    Written by Jay Kuri
    Copyright (c) 2006 Jay Kuri

    This work is licensed under the Creative Commons Attribution 2.5 License.
    To view a copy of this license, visit
    http://creativecommons.org/licenses/by/2.5/ or send a letter to Creative
    Commons, 543 Howard Street, 5th Floor, San Francisco, California, 94105,
    USA.
    
    
    Why a dragging controller when mootools already has a drag class?
    
    Quite simply, the Drag in mootools makes some assumptions that don't make
    sense for a lot of the things I needed to do. For example - I need to be able
    to turn on and off dragability.  I also needed to have access to the event
    objects in the on-event handlers (keyboard modifiers anyone?). 
    It also needed to not interfere with existing event handlers. Finally - 
    It made more sense to me to have something that controls the drag as an 
    effect, rather than making each element a new type of object. An instantiated 
    JKDrag object can handle dragging any element, all you need to do is call 
    your JKDrag's startDrag method with the element you want to start dragging. 
    This can be in the mousedown handler, the onclick handler, or in your own 
    code. Once the drag is complete, your dragged element is left as it was 
    before you started (unless you changed it in the on-event handlers, obviously)

*/

JKDrag = new Class({
    initialize: function(options) {
        this.options = {
                        onstart: null,
                        ondrop: null,
                        ondrag: null,
                        onclick: null,
                        hideoriginal: false,
                        ghostclass: 'JKDrag_ghost',
                        movethreshhold: 5,
                        hoverclass: 'JKDrag_dragoverhover',
                        draggingclass: 'JKDrag_dragging'
            };
        this.dragovers= new Array();
            
        if (options != null) {
            for (var option in options) {
                if (option == 'dragovers') {
                    this.dragovers = options[option];
                } else {
                    this.options[option] = options[option];
                }
            }
        }
        
        this.ghostelement = null;
        this.draggingelement = null;
        this.originaldisplay = null;
        this.currentlyover = null;
        this.clickargs = null;
        this.bound = new Object;
        this.options.movethreshhold = options.movethreshold || 5;
    },
    
    addDragover: function(element) {
        this.dragovers.push(element);
    },
    
    removeDragover: function(element) {
        var i = this.dragovers.indexOf(element);
        this.dragovers.splice(i,1);            
    },
    
    makeGhost: function(element) {
        var ghostelement = new Element('div');
        ghostelement.addClass(this.options.ghostclass);
        ghostelement.setStyle('position', 'absolute');
        var ghost = element.clone(true);
        ghostelement.adopt(ghost);
        return ghostelement;
    },
    
    startDrag: function(element, event) {
        var event = new Event(event);
        if (arguments.length > 2) {
            this.clickargs = new Array();
            for (var i = 2; i<arguments.length; i++) {
                this.clickargs.push(arguments[i]);
            }
        }

        this.moving = false;
        var allowdrag = true;
        if (this.options.onstart != null) {
            allowdrag = this.options.onstart(this, event, element);
        }
        if (allowdrag == true) {
            this.draggingelement = $(element);
            this.originalmouseposition = { x: event.page.x, y: event.page.y };
            this.originalposition = this.draggingelement.getPosition();
            this.bound.drag = this.drag.bindWithEvent(this);
    		this.bound.stopDrag = this.stopDrag.bindWithEvent(this);
    		document.addEvent('mousemove', this.bound.drag);
    		document.addEvent('mouseup', this.bound.stopDrag);
    	} else {
    	    this.originalmouseposition = null;
    	    this.originalposition = null;
    	    this.position = null;
    	}
    	event.stop();
    },
    
    isOver: function(location, element){
//		var elpos = $(element).getPosition();
//		return (location.left >= elpos.left && location.left <= elpos.right && 
//		        location.top <= elpos.bottom && location.top >= elpos.top);
		var elpos = this.getRealOffset(element);
		elpos.top = elpos.y;
		elpos.left = elpos.x;
		elpos.right = elpos.x + element.offsetWidth;
		elpos.bottom = elpos.y + element.offsetHeight;
		return (location.left >= elpos.left && location.left <= elpos.right && 
		        location.top <= elpos.bottom && location.top >= elpos.top);

	},
    
    getRealOffset: function (element) {
        var valueT = 0;
        var valueL = 0;
        var realParent = false;
        do {
            if (realParent) {
                while (realParent != element) {
                    valueT -= realParent.scrollTop  || 0;
                    valueL -= realParent.scrollLeft || 0;
                    realParent = realParent.parentNode;
                }
            }

            valueT += element.offsetTop  || 0;
            valueL += element.offsetLeft || 0;

            realParent = element.parentNode;
            element = element.offsetParent;
        } while (element);
        return { x: valueL, y: valueT };
    },
    
    getMouseOffset: function(event) {
        return { x: ((event.page.x - this.originalmouseposition.x)),
                 y: ((event.page.y - this.originalmouseposition.y)) };
    },
    
    drag: function(event) {
        var mouseoffset = this.getMouseOffset(event);
        if (this.moving == false) {
            var diffx = Math.abs(mouseoffset.x);
            var diffy = Math.abs(mouseoffset.y);
            if ((diffx+diffy) > this.options.movethreshhold ) {
                this.moving = true;
                this.ghostelement = this.makeGhost(this.draggingelement);
                if (this.options.hideoriginal == true) {
                    this.originaldisplay = this.draggingelement.getStyle('display');
                    this.draggingelement.setStyle('display', 'none');
                } else {
                    this.draggingelement.addClass(this.options.draggingclass);
                }
                if (this.options['ghostparentelement'] == null) {
                    this.ghostelement.injectAfter(this.draggingelement);
                } else {
                    this.options.ghostparentelement.adopt(this.ghostelement);
                }
                var baseoffset = this.getRealOffset(this.draggingelement);
                var ghostoffset = this.getRealOffset(this.ghostelement.offsetParent); 
                this.ghostoffset = { x: (baseoffset.x - ghostoffset.x),
                                     y: (baseoffset.y - ghostoffset.y)};
            }
        }
        if (this.moving == true) {
            //var originalpos = this.draggingelement.getPosition();
            var newpos = { x: (this.ghostoffset.x + mouseoffset.x),
                           y: (this.ghostoffset.y + mouseoffset.y)};
            var results = true; 
            if (this.options.ondrag != null) {
                results = this.options.ondrag(this, event, this.draggingelement, newpos);   
            }
            if (results == true) {
                results = { x: newpos.x, y: newpos.y };
            } else if (results == false) {
                results = { x: false, y: false };
            }
            if (results.y != false) {
                this.ghostelement.setStyle('top', results.y);
            }
            if (results.x != false) {
                this.ghostelement.setStyle('left', results.x);
            }
            var over = null;
            for (var i = 0; i < this.dragovers.length ; i++) {
                if (this.dragovers[i] != this.draggingelement) {
                    if (this.isOver(this.ghostelement.getPosition(), this.dragovers[i])) {
                        this.dragovers[i].addClass(this.options.hoverclass);
                        over = this.dragovers[i];
                    } else {
                        this.dragovers[i].removeClass(this.options.hoverclass);
                    }
                }
            }
            this.currentlyover = over;                   
            
        }
        event.stop();
    },
    
    stopDrag: function(event) {
        
        // always remove the events we added first - just in case one of the event handler 
        // callbacks fails, we don't wind up with an event that never gets detached.
        document.removeEvent('mousemove', this.bound.drag);
        document.removeEvent('mouseup', this.bound.stopDrag);
        var args = [event];
        if (this.clickargs != null) {
            args = args.concat(this.clickargs);
        }
        if (this.moving == false) {
            if (this.options.onclick != null) {
                this.options.onclick.apply(null, args);
            }
        } else {
            if (this.options.ondrop != null) {
                args.unshift(this);
                this.options.ondrop.apply(null, args);
            }
            if (this.draggingelement != null &&
                this.draggingelement.parentNode != null) {
            
                if (this.options.hideoriginal == true) {
                    this.draggingelement.setStyle('display', this.originaldisplay);
                } else {
                    this.draggingelement.removeClass(this.options.draggingclass);
                }
            }
            var i = 0;
            while (i < this.dragovers.length ) {
                this.dragovers[i].removeClass(this.options.hoverclass);
                // this check is to prevent leaks when dragover elements are removed from the dom
                // and not removed from our list.
                if (this.dragovers[i].parentNode == null) {
                    this.dragovers.splice(i,1);
                } else {
                    i++;
                }
            }
        }
        
        if (this.ghostelement != null) {
            this.ghostelement.remove();
            this.ghostelement = null;
        }
        this.bound.drag = null;
        this.bound.stopDrag = null;
        
        // let go of the original dragging element.
        this.draggingelement = null;
        // clear out the remaining per-drag variables;
        this.originaldisplay = null;    
        this.currentlyover = null;
        this.position = null;
        this.moving = false;
        this.clickargs = null;
        event.stop();
    }
});
